import logging

from django.apps import apps
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.utils import timezone

from .mixins import XAttrsMixin
from .utils import get_entity_pk

#from django.contrib.postgres.indexes import GinIndex

logger = logging.getLogger(__name__)


class Entity(XAttrsMixin, models.Model):
    """
    The central model of Onto, the Entity table contains a record for each record of every `EntityModel` subclass.

    Entities should NOT be created directly; they are automatically created whenever a new instance of an `EntityModel` subclass is created.
    """
    class Meta:
        verbose_name_plural = "entities"
#        indexes = [
#            GinIndex(fields=['xattrs']),
#            ]  # TODO Allow users to declare deeper indexes against xattrs

    class ArchiveManager(models.Manager):
        def from_object(self, obj) -> "Entity":
            """
            Return the primary Entity associated with any object, using its `entity_pk_field` attribute.

            EntityModel subclasses have this attribute by default, other classes must define it explicitly or this method will fail.
            """
            if isinstance(obj, Entity):
                return obj

            return self.get(pk=get_entity_pk(obj))

    class Manager(ArchiveManager):
        """
        Excludes archived entities.
        """
        def get_queryset(self):
            return super().get_queryset().exclude(archived_time__isnull=False)


    class QuerySet(models.QuerySet):
        def archive(self):
            for entity in self:
                entity.archive()

        def restore(self):
            for entity in self:
                entity.restore()

        def xfilter(self, **filters):
            """
            Include only entities with the specified extended attribute (`xattrs`).
            """
            return self.filter(**{f"xattrs__{query}":value for query,value in filters.items()})

        def as_model(self, *models: models.Model):
            """
            Returns a dictionary mapping Models to querysets of objects of that type.

            If exactly one model is specified, returns that queryset directly.

            Examples:

            >>> Entity.objects.filter(...).as_model(User)
            <QuerySet [<User: alice>, <User: bob>, ...]>

            >>> Entity.objects.filter(...).as_model(User, Region, ...)
            {
                User: <QuerySet [<User: alice>, <User: bob>, ...]>,
                Region: <QuerySet [<Region: fredonia>, ...]>,
                ...
            }
            """
            results = dict()
            models = models or {ct.model_class() for ct in ContentType.objects.filter(pk__in=self.values_list("content_type", flat=True).distinct())}
            pks = self.values("pk")
            for model in models:
                results[model] = model.objects.filter(**{f"{model.entity_pk_field}__in":pks})

            if len(models) == 1:
                return results[model]

            return results


    objects: Manager = Manager.from_queryset(QuerySet)()
    objects_archive: ArchiveManager = ArchiveManager.from_queryset(QuerySet)()

    entity_pk_field = "id"
    id = models.AutoField(primary_key=True)
    content_type = models.ForeignKey(ContentType, on_delete=models.RESTRICT, editable=False)
    content_object = GenericForeignKey('content_type', 'id')

    created_time = models.DateTimeField(
        auto_now_add=True,
        help_text="When this entity and associated object were created."
    )
    archived_time = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        db_index=True,
        help_text="When this entity and its associated object were last archived."
    )

    if apps.is_installed("onto.authz"):
        restricted = models.BooleanField(
            "Access restricted",
            default=False,
            blank=True,
            help_text="When False, access to this object may be granted simply by sharing a Domain. When True, there must also be a matching Policy.",
        )

    @property
    def url(self):
        """
        Return a relative link to the details page for this object.
        """
        # TODO add prefix from settings?
        return '/'.join(("", *self.content_type.natural_key(), str(self.id)))

    def archive(self):
        self.content_object.archived = True
        self.content_object.save(update_fields=["archived"])
        self.archived_time = timezone.now()
        return self.save(update_fields=["archived_time"])

    def related_objects(self, model) -> models.QuerySet:
        """
        Return a QuerySet of `model` where `model.entity_pk_field` resolves to this Entity.
        """
        return model.objects.filter(**{f"{model.entity_pk_field}":self.pk})

    def __str__(self):
        return '.'.join((*self.content_type.natural_key(), str(self.content_object)))

class EntityModel(models.Model):
    """
    Abstract class that overrides the default `id` primary key with a OneToOne field to the Entity table called `entity_id`.
    Also adds a `archived` BooleanField to allow "soft-deletion".

    This means each EntityModel represents two database records, one for the usual table, and one in the Entity table.
    This gives us the powerful ability to query all Entities at once with zero joins, even if the underlying schemas are incompatible.

    However, it does mean we must be careful with which models inherit this, so as to not pollute this precious global "namespace".
    Good candidates: Person, Location, Organization, Asset, System, Country, ~Article, etc. 
    (Entities or "business objects" that *exist* outside this database or would benefit strongly from the querying functionality.)
    Bad candidates: Event, Log, sensor reading, relationship/m2m tables, transaction record, etc.

    Otherwise, this EntityModel behaves like any other base Model subclass.
    """
    class Meta:
        abstract = True

    class QuerySet(models.QuerySet):
        @property
        def entities(self):
            """
            Return the QuerySet of corresponding Entities. This is a nested query, so could be slow for large QuerySets!
            """
            return Entity.objects_archive.filter(pk__in=self)

        def xfilter(self, **filters):
            """
            Include only entities with the specified extended attribute (`xattrs`).
            """
            return self.filter(**{f"entity__xattrs__{query}":value for query,value in filters.items()})

        def delete(self, *args, **kwargs):
            """
            Delete all objects in the QuerySet by deleting their associated Entities.
            """
            return self.entities.delete(*args, **kwargs)

        def archive(self):
            """
            "Soft-deletes" objects in the QuerySet, preventing them from showing up in the default `objects` Manager.

            Use the `objects_archive` Manager to query both archived and unarchived objects.
            """
            with transaction.atomic():
                self.entities.update(archived_time=timezone.now())
                self.update(archived=True)

        def restore(self):
            """
            Reverses the archiving i.e. "soft-deletion" of objects in the QuerySet (see `.archive()`).
            
            Note that the default `objects` Manager won't select archived objects, so this method will only be useful on the `objects_archive` Manager.
            """
            with transaction.atomic():
                self.update(archived=False)
                self.entities.update(archived_time=None)


    class ArchiveManager(models.Manager.from_queryset(QuerySet)):
        use_in_migrations = False  # Not worth the hassle/footguns

    class Manager(ArchiveManager):
        def get_queryset(self):
            return super().get_queryset().exclude(archived=True)

    objects = Manager()
    objects_archive = ArchiveManager()

    entity_pk_field = "entity_id"
    entity = models.OneToOneField(
        Entity,
        on_delete=models.CASCADE,
        primary_key=True,
        related_name="+",
        editable=False,
    )
    archived = models.BooleanField(
        default=False,
        editable=False,
        db_index=True,
        help_text="True indicates the object has been archived and won't appear in most queries."
    )

    @property
    def content_type(self):
        # This is cached by default.
        return ContentType.objects.get_for_model(type(self))

    def xget(self, key, default=None):
        """
        Get the associated entity's `xattrs` value by `key`, or `default` if it doesn't exist.
        """
        return self.entity.xget(key, default=default)

    def xset(self, **data) -> Entity:
        """
        Update the associated entity's `xattrs` with the provided keyword arguments.

        Don't forget to `xsave()`!
        """
        return self.entity.xset(**data)

    def xdel(self, *keys) -> Entity:
        """
        Delete the `keys` from the `xattrs` of the associated entity.

        Don't forget to `xsave()`!
        """
        return self.entity.xdel(*keys)

    def xsadd(self, **data) -> Entity:
        """
        Add to set. Attempt to add each of the keyword values to the sets at their respective keys in the associated entity's `xattrs`.

        Don't forget to `xsave()`!

        If a key does not exist, create a set there.

        If a key exists but is not a list, raise a ValueError.

        If the values are iterable (non-string), they *will* be unpacked/flattened!
        Example:
        >>> user.xsadd(roles="viewer")
        >>> user.xsadd(roles={"editor", "admin"})
        >>> user.xget("roles")
        ["viewer", "editor", "admin"]  # note how it is a flat list
        """
        return self.entity.xsadd(**data)

    def xsremove(self, **data) -> Entity:
        """
        Remove from set. Attempt to remove each of the keyword values from the sets at their respective keys in the associated entity's `xattrs`.

        Don't forget to `xsave()`!

        If a key exists but is not a list, raise a ValueError.

        If the values are iterable (non-string), they *will* be unpacked/flattened!
        """
        return self.entity.xsremove(**data)

    def xsave(self, *args, **kwargs):
        """
        `save` the `xattrs` field, committing it to the database.
        """
        return self.entity.xsave(*args, **kwargs)

    def save(self, *args, **kwargs):
        with transaction.atomic():
            created = self._state.adding
            if created:
                if not hasattr(self, "entity"):
                    ct = ContentType.objects.get_for_model(self.__class__)
                    self.entity = Entity.objects.create(content_type=ct)
            result = super().save(*args, **kwargs)
            if created:
                self.refresh_from_db()  # need this to make sure that (self.entity.content_object == self) right away
            return result

    def delete(self, *args, **kwargs):
        """
        Permanently deletes the object by deleting its associated Entity.
        """
        return self.entity.delete(*args, **kwargs)

    def archive(self):
        """
        Soft-deletes the object by marking it `archived` and preventing the default `objects` Manager from accessing it.
        """
        return self.entity.archive()

    def restore(self):
        """
        Reverses the archiving or "soft-deletion" of the object and its Entity.
        """
        return self.entity.restore()




from django.db import models


class XAttrsMixin(models.Model):
    """
    Adds an `xattrs` JSONField to the model, as well as supporting methods.

    IMPORTANT: After invoking one of these methods, make sure to `xsave()` to commit changes to the database!

    Make sure to inherit XAttrsMixin.QuerySet into the model's Manager to gain access to the custom methods like `xfilter`.
    """
    class Meta:
        abstract = True

    class QuerySet(models.QuerySet):
        def xfilter(self, **filters):
            """
            Include only objects with the specified extended attribute (`xattrs`).
            """
            return self.filter(**{f"xattrs__{query}":value for query,value in filters.items()})

    xattrs = models.JSONField(
        "extended attributes",
        blank=True,
        default=dict,
        help_text="Arbitrary JSON attributes, e.g. tags, to attach to this object."
    )

    def xget(self, key, default=None):
        """
        Get the object's `xattrs` value by `key`, or `default` if it doesn't exist.
        """
        return self.xattrs.get(key, default)

    def xset(self, **data):
        """
        Update the object's `xattrs` with the provided keyword arguments.

        Don't forget to `xsave()`!
        """
        for key, value in data.items():
            self.xattrs[key] = value
        return self

    def xdel(self, *keys) -> None:
        """
        Delete the `keys` from the `xattrs` of the object.

        Don't forget to `xsave()`!
        """
        for key in keys:
            try:
                del self.xattrs[key]
            except KeyError:
                pass
        return self

    def xsadd(self, **data):
        """
        Add to set. Attempt to add each of the keyword values to the sets at their respective keys.

        Don't forget to `xsave()`!

        If a key does not exist, create a set there.

        If a key exists but is not a list, raise a ValueError.

        If the values are iterable (non-string), they *will* be unpacked/flattened!
        Example:
        >>> user.xsadd(roles="viewer")
        >>> user.xsadd(roles={"editor", "admin"})
        >>> user.xget("roles")
        ["viewer", "editor", "admin"]  # note how it is a flat list
        """
        updated = dict()
        for key, value in data.items():
            elements = self.xget(key)
            if not elements:
                elements = value
            elif not isinstance(elements, list):
                raise ValueError(f"xattrs[key] exists, but it is not a list and cannot be added to!")
            else:
                if hasattr(value, "__iter__") and not isinstance(value, str):
                    for v in value:
                        if v not in elements:
                            elements.append(v)
                else:
                    if value not in elements:
                        elements.append(value)

            updated[key] = elements
        self.xset(**updated)
        return self

    def xsremove(self, **data):
        """
        Remove from set. Attempt to remove each of the keyword values from the sets at their respective keys.

        Don't forget to `xsave()`!

        If a key exists but is not a list, raise a ValueError.

        If the values are iterable (non-string), they *will* be unpacked/flattened!
        """
        updated = dict()
        for key, value in data.items():
            elements = self.xget(key)
            if not elements:
                return
            elif not isinstance(elements, list):
                raise ValueError(f"xattrs[key] exists, but it is not a list and cannot be deleted from!")
            else:
                if hasattr(value, "__iter__") and not isinstance(value, str):
                    for v in value:
                        elements.remove(v)
                else:
                    elements.remove(value)
            updated[key] = elements
        self.xset(**updated)
        return self

    def xsave(self, *args, **kwargs):
        """
        `save` the `xattrs` field, committing it to the database.
        """
        self.save(*args, update_fields=["xattrs"], **kwargs)
        return self


# Generated by Django 4.1.7 on 2023-03-16 20:06

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("onto", "0004_remove_entity_created_time_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="entity",
            name="created_time",
            field=models.DateTimeField(
                auto_now_add=True,
                default=django.utils.timezone.now,
                help_text="When this entity and associated object were created.",
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 4.1.7 on 2023-02-27 22:28

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Entity",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_time",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="When this entity and associated object were created.",
                    ),
                ),
                (
                    "updated_time",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="When this entity (but not *necessarily* its associated object) was last modified.",
                    ),
                ),
                (
                    "archived_time",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="When this entity and its associated object were last archived.",
                        null=True,
                    ),
                ),
                (
                    "xattrs",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        help_text="Arbitrary JSON attributes, e.g. tags.",
                        verbose_name="extended attributes",
                    ),
                ),
                (
                    "restricted",
                    models.BooleanField(
                        blank=True,
                        default=False,
                        help_text="When False, access to this object may be granted simply by sharing a Domain. When True, there must also be a matching Policy.",
                        verbose_name="Access restricted",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.RESTRICT,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "entities",
            },
        ),
        migrations.CreateModel(
            name="Domain",
            fields=[
                (
                    "entity",
                    models.OneToOneField(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="+",
                        serialize=False,
                        to="onto.entity",
                    ),
                ),
                (
                    "archived",
                    models.BooleanField(
                        db_index=True,
                        default=False,
                        editable=False,
                        help_text="True indicates the object has been archived and won't appear in most queries.",
                    ),
                ),
                ("name", models.SlugField(max_length=255, unique=True)),
                (
                    "authorizes",
                    models.BooleanField(
                        blank=True,
                        db_index=True,
                        default=False,
                        help_text="When True, membership to this Domain is considered for authorization. When False, membership will never affect authorization.",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Membership",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "xattrs",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        help_text="Arbitrary JSON attributes, e.g. tags, to attach to this object.",
                        verbose_name="extended attributes",
                    ),
                ),
                (
                    "entity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="memberships",
                        to="onto.entity",
                    ),
                ),
                (
                    "domain",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="memberships",
                        to="onto.domain",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="entity",
            name="domains",
            field=models.ManyToManyField(
                blank=True,
                help_text="Domains, or entity groups, that this object is a member of.",
                related_name="entities",
                through="onto.Membership",
                to="onto.domain",
            ),
        ),
        migrations.AddConstraint(
            model_name="membership",
            constraint=models.UniqueConstraint(
                fields=("entity", "domain"), name="onto_membership_unique"
            ),
        ),
    ]



# Generated by Django 4.1.7 on 2023-02-27 22:28

from django.db import migrations, models
import django.db.models.deletion
import onto.authz.models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("onto", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Action",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "label",
                    models.SlugField(
                        help_text="Internal reference key for this action.",
                        max_length=255,
                    ),
                ),
                (
                    "default_permit",
                    models.BooleanField(
                        default=True,
                        help_text="Whether this action is allowed to be performed by default. If False, a Policy must explicitly grant it.",
                    ),
                ),
                (
                    "dispatch_changes",
                    models.BooleanField(
                        default=False,
                        help_text="Whether to track the state of all entitlements for this action, and emit signals when they change.",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="A human-readable name for this action.",
                        max_length=255,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        help_text="Describe the action, including its intended purpose.",
                        null=True,
                    ),
                ),
                (
                    "source_type",
                    models.ForeignKey(
                        default=onto.authz.models.user_ct_id,
                        help_text="The ContentType of objects that perform this action, defaults to AUTH_USER_MODEL.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="actions_as_source",
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "target_type",
                    models.ForeignKey(
                        help_text="The ContentType of objects that this action affects.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="actions_as_target",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Policy",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("label", models.SlugField(max_length=255)),
                (
                    "principal_filters",
                    models.JSONField(
                        default=dict,
                        help_text="This policy affects entities (as source principals) that match this specification (or all entities in the domain if NULL).",
                    ),
                ),
                (
                    "resource_filters",
                    models.JSONField(
                        default=dict,
                        help_text="This policy affects entities (as target resources) that match this specification (or all entities in the domain if NULL).",
                    ),
                ),
                (
                    "disabled",
                    models.BooleanField(
                        default=False,
                        help_text="When this policy is disabled, it will not be considered for authorization.",
                    ),
                ),
                (
                    "actions",
                    models.ManyToManyField(
                        related_name="policies", to="onto_authz.action"
                    ),
                ),
                (
                    "domain",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="policies",
                        to="onto.domain",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "policies",
            },
        ),
        migrations.CreateModel(
            name="Entitlement",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "action",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="entitlements",
                        to="onto_authz.action",
                    ),
                ),
                (
                    "domain",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="entitlements",
                        to="onto.domain",
                    ),
                ),
                (
                    "policy",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="entitlements",
                        to="onto_authz.policy",
                    ),
                ),
                (
                    "principal",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="entitlements_as_principal",
                        to="onto.entity",
                    ),
                ),
                (
                    "resource",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="entitlements_as_resource",
                        to="onto.entity",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="policy",
            constraint=models.UniqueConstraint(
                fields=("label", "domain"), name="onto_authz_policy_unique"
            ),
        ),
        migrations.AddConstraint(
            model_name="action",
            constraint=models.UniqueConstraint(
                fields=("label", "target_type", "source_type"),
                name="onto_authz_action_unique",
            ),
        ),
    ]



# Generated by Django 4.1.7 on 2023-03-16 18:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("onto", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="domain",
            name="entity",
        ),
        migrations.RemoveField(
            model_name="membership",
            name="domain",
        ),
        migrations.RemoveField(
            model_name="membership",
            name="entity",
        ),
        migrations.RemoveField(
            model_name="entity",
            name="domains",
        ),
        migrations.AlterField(
            model_name="entity",
            name="xattrs",
            field=models.JSONField(
                blank=True,
                default=dict,
                help_text="Arbitrary JSON attributes, e.g. tags, to attach to this object.",
                verbose_name="extended attributes",
            ),
        ),
    ]



from django.contrib import admin

#from .indirect_inline import IndirectStackedInline
from django.contrib.contenttypes.admin import GenericStackedInline

from . import models

# Register your models here.

class EntityModelAdmin(admin.ModelAdmin):
    def get_inlines(self, request, obj):
        inlines = super().get_inlines(request, obj=obj)
        if EntityAdminInline not in inlines:
            return list(inlines) + [EntityAdminInline]
        return inlines

class EntityAdminInline(GenericStackedInline):
    model = models.Entity
    ct_fk_field = "id"
    extra = 3
    readonly_fields = ['id', 'content_type', 'domains', 'created_time', 'archived_time']
    fieldsets = (
        (None, {
            "fields": ('xattrs',)
        }),
        ("Timestamps", {
            "fields": ('created_time', 'archived_time'),
            "classes": ('collapse',),
        }),
        ("Advanced", {
            "fields": ('domains', 'content_type'),
            "classes": ('collapse',),
        })
    )

    def get_form_queryset(self, obj):
        return self.model.objects.filter(id=obj.entity_id)

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

@admin.register(models.Entity)
class EntityAdmin(admin.ModelAdmin):
    readonly_fields = ["content_type"]
    #inlines = [MembershipInline]



# Generated by Django 4.1.7 on 2023-03-16 18:43

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("onto_authz", "0002_domain_membership_domain_entities_and_more"),
        ("onto", "0002_remove_domain_entity_remove_membership_domain_and_more"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Domain",
        ),
        migrations.DeleteModel(
            name="Membership",
        ),
    ]



# Generated by Django 4.1.7 on 2023-03-16 20:05

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("onto", "0003_delete_domain_delete_membership"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="entity",
            name="created_time",
        ),
        migrations.RemoveField(
            model_name="entity",
            name="updated_time",
        ),
    ]



import logging
import re

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser, UserManager
from django.contrib.contenttypes.models import ContentType
from django.db import models

from onto.models import Entity, EntityModel, XAttrsMixin

from . import utils

audit_log = logging.getLogger("authz")

def user_ct_id():
    return ContentType.objects.get_for_model(get_user_model()).pk


# TODO add preferred language / timezone
class AbstractUser(EntityModel, AbstractUser):
    """
    Swaps `first_name`, `last_name` fields for more generic `display_name`.
    """
    class Meta:
        abstract = True

    class ArchiveManager(EntityModel.ArchiveManager, UserManager):
        pass

    class Manager(ArchiveManager):
        def get_queryset(self):
            return super().get_queryset().exclude(archived=True)


    objects = Manager()
    objects_archive = ArchiveManager()

    display_name = models.CharField(
        max_length=255,
        null=True,
        blank=True,
    )

    @property
    def first_name(self):
        if not self.display_name:
            return ""

        return self.display_name.split()[0]
        
    @property
    def last_name(self):
        if not self.display_name:
            return ""

        tokens = self.display_name.split()
        if len(tokens) <= 1:
            return ""

        return tokens[-1]

    # Technically, this is already monkeypatched onto EntityModel, but we also specify it here for a little more API clarity.
    def is_authorized(self, action, resource):
        """
        Returns True if the User is authorized to perform the `action` on the `resource`, otherwise False.
        """
        return is_authorized(self, action, resource)

    def __str__(self):
        return self.display_name or self.username or self.email


class Domain(models.Model):
    """
    A unique dot-delimited string representing a set of Entities.
    """
    class Manager(models.Manager):
        def get_by_natural_key(self, name):
            return self.get(name=name)

    objects = Manager()

    name = models.SlugField(
        unique=True,
        max_length=255,
        help_text="Fully-qualified Domain Name. Must be unique.",
    )

    entities = models.ManyToManyField(
        "onto.Entity",
        related_name="domains",
        blank=True,
        through="onto_authz.Membership",
        help_text="Entities contained in this domain."
    )

    active = models.BooleanField(
        default=True,
        help_text="When inactive, this domain cannot grant authorization."
    )

    def add(self, obj: EntityModel) -> "Membership":
        """
        Add an object to the domain, returning the associated Membership object.
        """
        membership, created = self.memberships.get_or_create(entity_id=obj.pk)
        return membership

    def get(self, obj: EntityModel) -> "Membership":
        """
        Return the Membership associated with an object. Raises DoesNotExist if not a member.
        """
        return self.memberships.get(entity_id=obj.pk)
    
    def remove(self, obj: EntityModel):
        """
        Remove an object from the domain, if it exists.
        """
        return self.memberships.filter(entity_id=obj.pk).delete()
    
    def xfilter(self, **filters) -> "Membership.QuerySet":
        return self.memberships.xfilter(**filters)

    @property
    def subdomains(self):
        return Domain.objects.filter(name__startswith=self.name+'.')

    def create_subdomain(self, subdomain_name, **kwargs) -> "Domain":
        """
        Creates a new domain with the current domain's name followed by `.subdomain_name`.

        Example:
        >>> domain1 = Domain.objects.create(name="acme")
        >>> domain2 = domain1.create_subdomain("accounting")
        >>> domain2.name
        "acme.accounting"
        >>> domain2 in domain1.subdomains
        True
        """
        subdomain, created = Domain.objects.get_or_create(name=f"{self.name}.{subdomain_name}", **kwargs)
        return subdomain

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return self.name
    
    def __contains__(self, item: EntityModel):
        return self.memberships.filter(entity_id=item.pk).exists()


class Membership(XAttrsMixin, models.Model):
    """
    Represents the Domain-Entity M2M relationship.

    Has its own set of `xattrs`, allowing arbitrary data to be associated with individual relationships.
    """
    class QuerySet(XAttrsMixin.QuerySet, models.QuerySet):
        @property
        def entities(self):
            return Entity.objects.filter(pk__in=self.values("entity").distinct())

        @property
        def domains(self):
            return Domain.objects.filter(pk__in=self.values("domain").distinct())

        def as_model(self, *models: models.Model, annotate_xattrs=False):
            """
            Returns a dictionary mapping Models to querysets of objects of that type, as contained in the Memberships.

            If exactly one model is specified, returns that queryset directly.

            If `annotate_xattrs` is True, annotate each object with its Membership `xattrs`; this uses a raw query for efficiency and cannot be further chained.
            """
            results = dict()
            models = models or {ct.model_class() for ct in ContentType.objects.filter(pk__in=self.values_list("content_type", flat=True).distinct())}
            for model in models:
                ct = ContentType.objects.get_for_model(model)
                matching = self.filter(entity__content_type=ct)  # Could remove the contenttype filter to remove a join at cost of larger IN, may want to profile
                if not annotate_xattrs:
                    results[model] = model.objects.filter(entity_id__in=matching.values("entity_id"))
                else:
                    results[model] = model.objects.raw(f"SELECT model.*, ms.xattrs::json FROM {model._meta.db_table} model INNER JOIN {Membership._meta.db_table} ms ON model.entity_id=ms.entity_id WHERE ms.id IN ({matching.values('id').query});")

            if len(models) == 1:
                return results[model]

            return results
        

    objects = QuerySet.as_manager()

    domain = models.ForeignKey(
        Domain,
        on_delete=models.CASCADE,
        related_name="memberships",
    )
    entity = models.ForeignKey(
        Entity,
        on_delete=models.CASCADE,
        related_name="memberships",
    )

    @property
    def object(self):
        return self.entity.content_object

    def __str__(self):
        return f"{self.entity}@{self.domain}"

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["entity", "domain"], name="%(app_label)s_%(class)s_unique")
        ]

class Action(models.Model):
    """
    An action that can be authorized. Does not specify what the action does.
    """
    class Manager(models.Manager):
        def get_by_natural_key(self, source_type_nk, label, target_type_nk):
            source_type = ContentType.objects.get_by_natural_key(*source_type_nk.split('.'))
            target_type = ContentType.objects.get_by_natural_key(*target_type_nk.split('.'))
            return self.get(label=label, source_type=source_type, target_type=target_type)

        def get_by_str(self, string):
            """
            Deserializes a __str__ into an Action object using the natural key. 
            """
            tokens = string.split('__')
            if len(tokens) == 2:
                tokens = (settings.AUTH_USER_MODEL.lower(), *tokens)
            return self.get_by_natural_key(*tokens)

    objects = Manager()

    label = models.SlugField(
        max_length=255,
        help_text="Internal reference key for this action.",
    )
    source_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        default=user_ct_id,
        related_name='actions_as_source',
        help_text="The ContentType of objects that perform this action, defaults to AUTH_USER_MODEL."
    )
    target_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        related_name='actions_as_target',
        help_text="The ContentType of objects that this action affects."
    )
    default_permit = models.BooleanField(
        default=True,
        help_text="Whether this action is allowed to be performed by default. If False, a Policy must explicitly grant it."
    )
    dispatch_changes = models.BooleanField(
        default=False,
        help_text="Whether to track the state of all entitlements for this action, and emit signals when they change."
    )

    title = models.CharField(
        max_length=255,
        help_text="A human-readable name for this action."
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text="Describe the action, including its intended purpose."
    )

    def natural_key(self):
        return (
        '.'.join(self.source_type.natural_key()),
        self.label,
        '.'.join(self.target_type.natural_key()),
        )

    def save(self, *args, **kwargs):
        self.label = re.sub(r'_{2,}','_', self.label)  # collapse underscores
        super().save(*args, **kwargs)
        
    def __str__(self):
        return "__".join(self.natural_key())

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["label", "target_type", "source_type"], name="%(app_label)s_%(class)s_unique")
        ]


class Policy(models.Model):
    """
    A Policy defines sets of valid Principals, Actions, and Resources, respectively.
    A valid Principal is authorized to perform any of the valid Actions on any of the valid Resources.
    """
    class Manager(models.Manager):
        def create_with_actions(self, *actions, **kwargs):
            policy = self.create(**kwargs)
            for action in actions:
                if isinstance(action, str):
                    action = Action.objects.get_by_str(action)
                policy.actions.add(action)

            return policy

    class QuerySet(models.QuerySet):
        pass

    objects = Manager.from_queryset(QuerySet)()

    domain = models.ForeignKey(
        Domain,
        on_delete=models.CASCADE,
        db_index=True,
        editable=False,
        related_name="policies",
    )
    label = models.SlugField(
        max_length=255,
    )

    principal_filters = models.JSONField(
        default=dict,
        help_text="This policy affects entities (as source principals) that match this specification (or all entities in the domain if NULL).",
    )
    actions = models.ManyToManyField(
        Action,
        related_name="policies",
    )
    resource_filters = models.JSONField(
        default=dict,
        help_text="This policy affects entities (as target resources) that match this specification (or all entities in the domain if NULL).",
    )
    disabled = models.BooleanField(
        default=False,
        help_text="When this policy is disabled, it will not be considered for authorization."
    )

    @property
    def principals(self) -> Entity.QuerySet:
        """
        All principal Entities that this Policy currently applies to.
        """
        return self.domain.entities.filter(content_type__in=self.actions.values("source_type").distinct(), **self.principal_filters)

    @property
    def resources(self) -> Entity.QuerySet:
        """
        All target Entities (i.e. resources) that this Policy currently applies to.
        """
        return self.domain.entities.filter(content_type__in=self.actions.values("target_type").distinct(), **self.resource_filters)

    def __str__(self):
        return f"{self.domain}.{self.label}"


    class Meta:
        verbose_name_plural = "policies"
        constraints = [
            models.UniqueConstraint(fields=["label", "domain"], name="%(app_label)s_%(class)s_unique")
        ]


class Entitlement(models.Model):
    """
    Entitlements track the state of a particular authorized action. They are managed automatically and should not be modified manually.
    See "./signals.py" for more details on the automated Entitlement system.
    """
    principal = models.ForeignKey(
        "onto.Entity",
        on_delete=models.CASCADE,
        editable=False,
        related_name="entitlements_as_principal",
    )
    action = models.ForeignKey(
        Action,
        on_delete=models.CASCADE, 
        editable=False,
        related_name="entitlements",
    )
    resource = models.ForeignKey(
        "onto.Entity",
        on_delete=models.CASCADE,
        editable=False,
        related_name="entitlements_as_resource",
    )
    domain = models.ForeignKey(
        "onto_authz.Domain",
        on_delete=models.CASCADE,
        editable=False,
        related_name="entitlements",
    )
    policy = models.ForeignKey(
        Policy,
        on_delete=models.CASCADE,
        null=True,
        editable=False,
        related_name="entitlements",
    )

    @property
    def peers(self):
        return Entitlement.objects.filter(principal=self.principal, resource=self.resource, action=self.action).exclude(id=self.id)

@utils.monkeypatch_method(EntityModel)
def is_authorized(principal: Entity, action: Action, resource: Entity):
    """
    Returns True iff the `principal` Entity (or EntityModel subclass) is allowed to perform `action` on `resource`, based on Domain Membership, Policies, and other factors.
    """

    principal = Entity.objects.from_object(principal)
    resource = Entity.objects.from_object(resource)

    if isinstance(action, str):
        action = Action.objects.get(label=action, source_type=principal.content_type, target_type=resource.content_type)
    elif action.source_type != principal.content_type or action.target_type != resource.content_type:
        raise ValueError(f"the action `{action}` does not match the provided principal's and resources' content_types.")

    intersection = set(principal.domains.filter(active=True) & resource.domains.filter(active=True))
    if not intersection:  # Entities MUST share a domain to interact.
        audit_log.info(f"DENIED {principal} {action.label} {resource} (no shared domain with active=True)")
        return False

    if action.default_permit and not resource.restricted:  # If the action is not access-controlled, and they entities share a domain, access is granted.
        audit_log.info(f"ALLOWED {principal} {action.label} {resource} (shared {intersection})")
        return True

    # For restricted actions, we must search applicable policies
    for policy in Policy.objects.filter(
        disabled=False,
        domain__in=intersection,
        actions=action,
    ):
        if (
            policy.principals.filter(pk=principal.pk).exists() 
            and policy.resources.filter(pk=resource.pk).exists()
        ):
            audit_log.info(f"ALLOWED {principal} {action.label} {resource} (matched policy {policy})")
            return True
    
    audit_log.info(f"DENIED {principal} {action.label} {resource} (no policy match)")
    return False

@utils.monkeypatch_method(EntityModel)
def get_authorized_resources(principal: Entity, action: Action, resource_model: models.base.ModelBase):
    """
    Returns a QuerySet of `resource_model` containing all objects that `principal` is currently authorized to perform `action` on.

    For example, answers the question "which Articles can this User 'edit'?"
    """

    principal = Entity.objects.from_object(principal)

    # NOTE: this does not currently handle EntityModel subclasses that have overriden `entity_pk_field`
    if issubclass(resource_model, EntityModel):
        resource_ct = ContentType.objects.get_for_model(resource_model)
    else:
        resource_ct = ContentType.objects.get_for_model(resource_model._meta.get_field(resource_model.entity_pk_field).remote_field.model)

    if isinstance(action, str):
        action = Action.objects.get(label=action, source_type=principal.content_type, target_type=resource_ct)

    if action.default_permit:
        resource_entities = Entity.objects.filter(restricted=False, domains__in=principal.domains.filter(active=True))
    else:
        resource_entities = Entity.objects.none()

    for policy in Policy.objects.filter(
      disabled=False,
      domain__in=principal.domains.filter(active=True),
      actions=action
      ):
        if not policy.principals.filter(pk=principal.pk).exists():
            continue
        resource_entities = resource_entities.union(policy.resources.values("pk"))

    return resource_entities.as_model(resource_model)

@utils.monkeypatch_method(EntityModel)
def get_authorized_principals(resource: Entity, action: Action, principal_model: models.base.ModelBase):
    """
    Returns a QuerySet of `principal_model` containing all objects that are currently authorized to perform `action` on `resource`.

    For example, answers the question "which Users can 'edit' this Article?"
    """

    resource = Entity.objects.from_object(resource)

    # NOTE: this does not currently handle EntityModel subclasses that have overriden `entity_pk_field`
    if issubclass(principal_model, EntityModel):
        principal_ct = ContentType.objects.get_for_model(principal_model)
    else:
        principal_ct = ContentType.objects.get_for_model(principal_model._meta.get_field(principal_model.entity_pk_field).remote_field.model)

    if isinstance(action, str):
        action = Action.objects.get(label=action, source_type=principal_ct, target_type=resource.content_type)

    if action.default_permit:
        principal_entities = Entity.objects.filter(domains__in=resource.domains.filter(active=True))
    else:
        principal_entities = Entity.objects.none()

    for policy in Policy.objects.filter(
      disabled=False,
      domain__in=resource.domains.filter(active=True),
      actions=action
      ):
        if not policy.resources.filter(pk=resource.pk).exists():
            continue
        principal_entities = principal_entities.union(policy.principals.values("pk"))

    return principal_entities.as_model(principal_model)


from onto.models import Entity, EntityModel

def obj_to_entity(obj):
    if isinstance(obj, EntityModel):
        return obj.entity
    elif isinstance(obj, Entity):
        return obj
    #elif hasattr(obj, "authz_proxy"):
    #    return obj_to_entity(getattr(obj, obj.authz_proxy))
    else:
        return None

# Snippet from GvR: https://mail.python.org/pipermail/python-dev/2008-January/076194.html
def monkeypatch_method(cls):
    def decorator(func):
        setattr(cls, func.__name__, func)
        return func
    return decorator


from django.apps import AppConfig


class ontoAuthConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'onto.authz'
    label = 'onto_authz'

    def ready(self) -> None:
        super().ready()
        from . import signals


import re

from django.utils.text import slugify


def get_entity_pk(obj):
    """
    Return the PK of the 'primary' Entity associated with this object, based on its `entity_pk_field` attribute.
    """
    try:
        return getattr(obj, obj.entity_pk_field)
    except AttributeError as e:
        raise AttributeError(f"{repr(obj)} cannot be resolved to an Entity because it is missing an 'entity_pk_field' attribute.") from e


from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as DjangoBaseUserAdmin

from onto.admin import EntityModelAdmin

from . import models

# Register your models here.

class BaseUserAdmin(EntityModelAdmin, DjangoBaseUserAdmin):
    # Swap first_name and last_name with display_name
    def get_list_display(self, request):
        list_display = list(super().get_list_display(request))
        list_display.insert(list_display.index("first_name"), "display_name")
        list_display.remove("first_name")
        list_display.remove("last_name")
        return list_display

    def get_fieldsets(self, request, obj):
        fieldsets = super().get_fieldsets(request, obj)
        for section, content in fieldsets:
            if section == "Personal info":
                content["fields"] = ("display_name", "email")
        return fieldsets
        
@admin.register(models.Policy)
class PolicyAdmin(admin.ModelAdmin):
    filter_horizontal = ["actions"]

@admin.register(models.Domain)
class DomainAdmin(EntityModelAdmin):
    filter_horizontal = ["entities"]
    search_fields = ['name']


"""
onto Authz uses many (optional) signals to automate Entitlement management. This is crucial for keeping distributed authorization systems in sync.

For example, changing the `xattrs` or `domains` of an Entity may change which Policies apply to it, which may need to be communicated to a remote authorization system.
This is all automated here, more or less with running as few queries as possible while keeping a distributed system consistent!
"""
from collections import namedtuple
from functools import cache

from django.db.models import Q, signals
from django.dispatch import Signal, receiver

import onto.authz.models
import onto.models

audit_log = onto.authz.models.audit_log

entitlement_enabled = Signal()
entitlement_disabled = Signal()

class silent_entitlements:
    """
    Context manager that suppresses `entitlement_enabled` and `entitlement_disabled` signals from firing within the scope.

    Primarily useful for "reverse-synchronizing" the database with a remote system, as a one-off or infrequent procedure.

    >>> with silent_entitlements():
    ...     # This won't cause any of the above signals to fire
    ...     something_that_affects_entitlements()
    
    """
    active = False
    def __enter__(self):
        silent_entitlements.active = True
    def __exit__(self, type, value, traceback):
        silent_entitlements.active = False


@receiver(signals.post_save, sender=onto.authz.models.Entitlement)
def on_entitlement_save(instance, sender,**kwargs):
    """
    When an Entitlement is saved, if its the only one of its kind, emit an `entitlement_enabled` signal.
    """
    print("on_entitlement_save", instance)
    audit_log.info(f"ENTITLEMENT ADDED {instance.principal} {instance.action.label} {instance.resource} (domain: {instance.domain}, policy: {instance.policy})")

    if not silent_entitlements.active and not instance.peers.exists():  # TODO Verify this won't race.
        audit_log.info(f"ENABLED {instance.principal} {instance.action.label} {instance.resource} (first entitlement gained)")
        entitlement_enabled.send(
            sender=sender,
            principal=instance.principal,
            action=instance.action,
            resource=instance.resource,
            )

@receiver(signals.post_delete, sender=onto.authz.models.Entitlement)
def on_entitlement_delete(instance, sender, **kwargs):
    """
    When an Entitlement is deleted, if it was the only one of its kind, emit an `entitlement_disabled` signal.
    """
    print("on_entitlement_delete", instance)
    audit_log.info(f"ENTITLEMENT REMOVED {instance.principal} {instance.action.label} {instance.resource} (domain: {instance.domain}, policy: {instance.policy})")

    if not silent_entitlements.active and not instance.peers.exists():
        audit_log.info(f"DISABLED {instance.principal} {instance.action.label} {instance.resource} (no entitlements remaining)")
        entitlement_disabled.send(
            sender=sender,
            principal=instance.principal,
            action=instance.action,
            resource=instance.resource,
            )

@receiver(signals.post_save, sender=onto.authz.models.Action)
def on_action_save(**kwargs):
    get_dispatching_actions.cache_clear()
    get_affected_types.cache_clear()

@cache
def get_dispatching_actions():
    return set(onto.authz.models.Action.objects.filter(dispatch_changes=True))

@cache
def get_affected_types():
    actions = get_dispatching_actions()
    return {action.source_type for action in actions}.union({action.target_type for action in actions})

#
# Below are various receivers to support the automation of Entitlement management
#
Policies = namedtuple('Policies', ('as_principal', 'as_resource'))

def evaluate_policies(instance: onto.models.Entity, past_policies: Policies, domains=None):
    print("evaluate_policies", instance)
    # A user or access-controlled Entity had changes to its security attributes, so reevaluate Policies.
    # First recompile Entitlements where the instance is a Principal
    domains = domains or instance.domains.filter(active=True)
    potential_policies = onto.authz.models.Policy.objects.filter(domain__in=domains, actions__source_type=instance.content_type)
    matching_policies = {policy for policy in potential_policies if policy.principals.filter(pk=instance.pk).exists()}
    previous_matching_policies = set(past_policies.as_principal)

    for new_policy in matching_policies - previous_matching_policies:
        # create entitlements with instance as target resource
        for action in new_policy.actions.filter(source_type=instance.content_type):
            for resource in new_policy.resources:
                action.entitlements.create(
                    domain=new_policy.domain,
                    policy=new_policy,
                    principal=instance,
                    resource=resource,
                )
    for old_policy in previous_matching_policies - matching_policies:
        old_policy.entitlements.filter(principal=instance).delete()

    # Now recompile Entitlements where the instance is a target Resource
    potential_policies = onto.authz.models.Policy.objects.filter(domain__in=instance.domains.filter(active=True), actions__target_type=instance.content_type)
    matching_policies = {policy for policy in potential_policies if policy.resources.filter(pk=instance.pk).exists()}
    previous_matching_policies = set(past_policies.as_resource)

    for new_policy in matching_policies - previous_matching_policies:
        # create entitlements with instance as target resource
        for action in new_policy.actions.filter(target_type=instance.content_type):
            for principal in new_policy.principals:
                action.entitlements.create(
                    domain=new_policy.domain,
                    policy=new_policy,
                    principal=principal,
                    resource=instance,
                )
    for old_policy in previous_matching_policies - matching_policies:
        old_policy.entitlements.filter(resource=instance).delete()


@receiver(signals.pre_save, sender=onto.models.Entity)
def before_entity_save(instance, sender, **kwargs):
    """
    Before an Entity is saved, check if its `xattrs` was modified, so we can update Entitlements later.
    """
    print("before_entity_save", instance)
    if instance.content_type not in get_affected_types():
        return
    try:
        db_instance = sender.objects_archive.get(pk=instance.pk)
        if instance.xattrs != db_instance.xattrs:
            instance._db_policies = Policies(
                as_principal=onto.authz.models.Policy.objects.filter(entitlements__principal=db_instance).distinct(),
                as_resource=onto.authz.models.Policy.objects.filter(entitlements__resource=db_instance).distinct(),
            )
    except sender.DoesNotExist:  # New Entity
        instance._db_policies = Policies(
            as_principal=onto.authz.models.Policy.objects.none(),
            as_resource=onto.authz.models.Policy.objects.none(),
        )


@receiver(signals.post_save, sender=onto.models.Entity)
def after_entity_save(instance, sender, **kwargs):
    """
    After an Entity is saved, if its security attributes were modified, create/delete Entitlements accordingly.
    """
    print("after_entity_save", instance)
    if getattr(instance, "_db_policies", None):
        evaluate_policies(instance, instance._db_policies)
        delattr(instance, "_db_policies")


def on_membership_added(domain, entity):
    print("on_membership_added", domain, entity)
    if not domain.active:
        return

    actions = onto.authz.models.Action.objects.filter(dispatch_changes=True)

    # Create principal entitlements for any dispatching actions that apply to this entity
    for action in actions.filter(source_type=entity.content_type):
        for resource in domain.entities.filter(content_type=action.target_type, restricted=False):
            action.entitlements.create(
                domain=domain,
                principal=entity,
                resource=resource,
                policy=None,
            )
    # Create target resource entitlements for any dispatching actions that apply to this entity
    for action in actions.filter(target_type=entity.content_type):
        for principal in domain.entities.filter(content_type=action.source_type, restricted=False):
            action.entitlements.create(
                domain=domain,
                principal=principal,
                resource=entity,
                policy=None,
            )

def on_membership_removed(domain, entity):
    print("on_membership_removed", domain, entity)
    domain.entitlements.filter(Q(principal=entity)|Q(resource=entity)).delete()

@receiver(signals.pre_save, sender=onto.authz.models.Membership)
def before_membership_save(instance, sender, **kwargs):
    """
    Before a Membership is saved, check if its `xattrs` was modified, so we can update the associated Entity's Entitlements later.
    """
    print("before_membership_save", instance)
    try:
        db_instance = sender.objects.get(pk=instance.pk)
        if instance.xattrs == db_instance.xattrs:
            return
    except sender.DoesNotExist:  # New Membership -- handled by `after_membership_save`
        pass

    instance.entity._db_policies = Policies(
        as_principal=onto.authz.models.Policy.objects.filter(entitlements__principal=instance.entity).distinct(),
        as_resource=onto.authz.models.Policy.objects.filter(entitlements__resource=instance.entity).distinct(),
    )

@receiver(signals.post_save, sender=onto.authz.models.Membership)
def after_membership_save(instance, sender, created, **kwargs):
    """
    After an Membership is saved, if its extended attributes were modified, create/delete Entitlements accordingly for its associated Entity.
    """
    print("after_membership_save", instance)
    domain = instance.domain
    entity = instance.entity

    if created:
        on_membership_added(domain, entity)

    if getattr(entity, "_db_policies", None):
        evaluate_policies(entity, entity._db_policies, domains=(domain,))
        delattr(entity, "_db_policies")


@receiver(signals.post_delete, sender=onto.authz.models.Membership)
def after_membership_delete(instance, sender, **kwargs):
    """
    After an Membership is saved, clean up any corresponding Entitlements.
    """
    print("after_membership_delete", instance)
    on_membership_removed(instance.domain, instance.entity)

@receiver(signals.pre_save, sender=onto.authz.models.Policy)
def before_policy_save(instance, sender, **kwargs):
    """
    Before a Policy is changed, identify any changes in the filters.
    """
    print("before_policy_save", instance)
    try:
        db_instance = sender.objects.get(pk=instance.pk)
        if instance.principal_filters != db_instance.principal_filters:
            instance._db_principals = set(db_instance.principals)
        if instance.resource_filters != db_instance.resource_filters:
            instance._db_resources = set(db_instance.resources)
    except sender.DoesNotExist:
        instance._db_principals = set()
        instance._db_resources = set()


@receiver(signals.post_save, sender=onto.authz.models.Policy)
def after_policy_save(instance, created, sender, **kwargs):
    """
    After a Policy is saved, if its filters were changed, adjust associated Entitlements accordingly.
    """
    print("after_policy_save", instance)
    if not created and (instance.disabled or not instance.domain.active):
        return instance.entitlements.all().delete()

    if getattr(instance, "_db_principals", False):
        # Principal filters were changed
        principals = set(instance.principals)
        for new_principal in principals - instance._db_principals:
            for action in instance.actions.filter(source_type=new_principal.content_type):
                for resource in instance.resources:
                    action.entitlements.create(
                        domain=instance.domain,
                        policy=instance,
                        principal=new_principal,
                        resource=resource,
                    )
        for old_principal in instance._db_principals - principals:
            instance.entitlements.filter(principal=old_principal).delete()
        delattr(instance, "_db_resources")

    if getattr(instance, "_db_resources", False):
        resources = set(instance.resources)
        for new_resource in resources - instance._db_resources:
            for action in instance.actions.filter(target_type=new_resource.content_type):
                for principal in instance.principals:
                    action.entitlements.create(
                        domain=instance.domain,
                        policy=instance,
                        principal=principal,
                        resource=new_resource,
                    )
        for old_resource in instance._db_resources - resources:
            instance.entitlements.filter(resource=old_resource).delete()
        delattr(instance, "_db_resources")


@receiver(signals.m2m_changed, sender=onto.authz.models.Policy.actions.through)
def on_policy_action_change(action, instance, pk_set, **kwargs):
    print("on_policy_action_change", instance)
    change_type = action; del action  # 'action' param is Django's not ours, so let's fix that for clarity.

    if change_type not in ("post_add", "post_remove",):  # post_clear?
        return
    if isinstance(instance, onto.authz.models.Policy):
        policies = [instance]
        actions = onto.authz.models.Action.objects.filter(pk__in=pk_set)
    else:
        actions = [instance]
        policies = onto.authz.models.Policy.objects.filter(pk__in=pk_set)

    for policy in policies:
        if change_type == "post_remove":
            policy.entitlements.filter(action__in=actions).delete()
        elif change_type == "post_add":
            for action in actions:
                for resource in policy.resources.filter(content_type=action.target_type):
                    for principal in policy.principals.filter(content_type=action.source_type):
                        action.entitlements.create(
                            domain=policy.domain,
                            policy=policy,
                            principal=principal,
                            resource=resource,
                        )


@receiver(signals.m2m_changed, sender=onto.models.Entity.domains.through)
def on_domain_entities_change(action, instance, pk_set, **kwargs):
    """
    When an Entity is added or removed from a Domain, we create or delete Entitlements accordingly.
    """
    print("on_domain_entities_change", instance)
    change_type = action; del action  # 'action' param is Django's not ours, so let's fix that for clarity.
    if change_type not in ("post_add", "post_remove",):  # post_clear?
        return

    if isinstance(instance, onto.authz.models.Domain):
        domains = [instance] if instance.active else []
        entities = onto.models.Entity.objects.filter(pk__in=pk_set, content_type__in=get_affected_types())
    elif instance.content_type in get_affected_types():
        entities = [instance]
        domains = onto.authz.models.Domain.objects.filter(pk__in=pk_set, active=True)
    else:
        return

    for domain in domains:
        for entity in entities:
            if change_type == "post_add":
                on_membership_added(domain, entity)

            elif change_type == "post_remove":
                on_membership_removed(domain, entity)

@receiver(signals.post_save, sender=onto.authz.models.Domain)
def on_domain_saved(instance, created, **kwargs):
    print("on_domain_saved", instance)
    if not instance.active and not created:
        instance.entitlements.all().delete()


# Generated by Django 4.1.7 on 2023-03-16 18:51

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("onto_authz", "0002_domain_membership_domain_entities_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="domain",
            name="active",
            field=models.BooleanField(
                default=True,
                help_text="When inactive, this domain cannot grant authorization.",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['authz',
 'authz.migrations',
 'onto',
 'onto.authz',
 'onto.authz.migrations',
 'onto.migrations']

package_data = \
{'': ['*']}

install_requires = \
['django>=4.1.6,<5.0.0']

setup_kwargs = {
    'name': 'django-onto',
    'version': '0.1.6a0',
    'description': "Onto (as in 'ontology') is a lightweight extension to Django centered around the concept of an Entity graph, which can be used to simplify the implementation of complex business logic. Onto includes a number of subpackages that leverage the Entity graph, including the powerful `onto.authz` authorization plugin.",
    'long_description': 'Placeholder - Onto source is being migrated to this new destination.\n',
    'author': 'Paul Fischer',
    'author_email': 'pphysch@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# Generated by Django 4.1.7 on 2023-03-16 18:43

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("onto", "0002_remove_domain_entity_remove_membership_domain_and_more"),
        ("onto_authz", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Domain",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.SlugField(
                        help_text="Fully-qualified Domain Name. Must be unique.",
                        max_length=255,
                        unique=True,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Membership",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "xattrs",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        help_text="Arbitrary JSON attributes, e.g. tags, to attach to this object.",
                        verbose_name="extended attributes",
                    ),
                ),
                (
                    "domain",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="memberships",
                        to="onto_authz.domain",
                    ),
                ),
                (
                    "entity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="memberships",
                        to="onto.entity",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="domain",
            name="entities",
            field=models.ManyToManyField(
                blank=True,
                help_text="Entities contained in this domain.",
                related_name="domains",
                through="onto_authz.Membership",
                to="onto.entity",
            ),
        ),
        migrations.AlterField(
            model_name="entitlement",
            name="domain",
            field=models.ForeignKey(
                editable=False,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="entitlements",
                to="onto_authz.domain",
            ),
        ),
        migrations.AlterField(
            model_name="policy",
            name="domain",
            field=models.ForeignKey(
                editable=False,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="policies",
                to="onto_authz.domain",
            ),
        ),
        migrations.AddConstraint(
            model_name="membership",
            constraint=models.UniqueConstraint(
                fields=("entity", "domain"), name="onto_authz_membership_unique"
            ),
        ),
    ]
